import asyncio

from .messages import ChannelEvents


class ChannelJoinFailure(Exception):
    pass


class ChannelLeaveFailure(Exception):
    pass

# TODO: Random thought, but _might_ be nice to ditch the
# mutable-ness of these classes.
# Like a Channel can be joined or not.
# Maybe we should have a JoinedChannel class vs a Channel class.
# `.channel` always returns the Channel.
# `.join` returns the JoinedChannel (and if already joined does nothing extra).
# Not sure if it'd make a good API, but worth thinking about...


class Channel:
    '''
    A channel on a phoenix server.

    Should not be instantiated directly, but through a socket.
    '''
    def __init__(self, socket, topic, params):
        self.socket = socket
        self.topic = topic
        self.params = params
        self._incoming_messages = asyncio.Queue()
        # TODO: Consider something like channel_states in js lib?

    async def join(self):
        '''
        Joins the channel.
        '''
        join = await self.socket._send_message(
            self.topic, ChannelEvents.join.value, self.params
        )
        try:
            response = await join.response()
        except Exception as e:
            # TODO: this needs some work.
            raise ChannelJoinFailure() from e

        return response

    async def leave(self):
        '''
        Leaves the channel.
        '''
        leave = await self.socket._send_message(
            self.topic, ChannelEvents.leave.value, self.params
        )
        try:
            response = await leave.response()
        except Exception as e:
            # TODO: this needs some work.
            raise ChannelLeaveFailure() from e

    async def push(self, event, payload):
        '''
        Pushes a message to a channel.

        :param event:    The event to push.
        :param payload:  The payload for the event.
        '''
        msg = await self.socket._send_message(self.topic, event, payload)
        return msg

    # TODO: could be nice to just expose a "read only queue" under .incoming
    # With get, get_nowait & an async iterator interface?
    # TODO: Otherwise should maybe be called pull (to go with push)
    async def receive(self):
        msg = await self._incoming_messages.get()
        return msg

    async def __aenter__(self):
        resp = await self.join()
        return self, resp

    async def __aexit__(self, exc_type, exc_value, tb):
        await self.leave()



from urllib.parse import urlencode
import asyncio
import json
import logging

import websockets

from .base import BaseTransport, TransportMessage
from ..utils import get_unless_done, DONE

__all__ = ['WebsocketTransport']

logger = logging.getLogger(__name__)


class WebsocketTransport(BaseTransport):
    '''
    Implements the websocket transport for talking to phoenix servers.
    '''
    def __init__(self, url, params, incoming_queue, outgoing_queue):
        super().__init__(
            incoming_queue=incoming_queue, outgoing_queue=outgoing_queue
        )
        qs_params = {'vsn': '1.0.0', **params}
        self.url = url + '?' + urlencode(qs_params)
        print(self.url)
        self.ready = asyncio.Future()
        self._done = asyncio.Future()

    async def run(self):
        try:
            async with websockets.connect(self.url) as websocket:
                self.ready.set_result(True)
                # TODO: Think about error propagation at some point?
                # If one of these crashes, does it cancel the other?
                # It should, not sure if it does.
                await asyncio.gather(
                    self._recv_loop(websocket), self._send_loop(websocket)
                )
        except Exception as e:
            if not self.ready.done():
                self.ready.set_exception(e)

            raise

    async def stop(self):
        self._done.set_result(True)

    async def _recv_loop(self, websocket):
        while True:
            message_data = await get_unless_done(websocket.recv(), self._done)
            if message_data is DONE:
                return

            logger.info("received: %s", message_data)
            # TODO: This needs updates.
            message = _load_incoming_message(json.loads(message_data))
            await self.incoming.put(message)
            logger.info("sent")

    async def _send_loop(self, websocket):
        while True:
            message = await get_unless_done(self.outgoing.get(), self._done)
            if message is DONE:
                return

            logger.info("sending: %s", message)
            try:
                # TODO: This needs updates.
                message_data = json.dumps(
                    dump_outgoing_message(message.message)
                )
                await websocket.send(message_data)
                message.sent.set_result(True)
            except Exception as e:
                message.sent.set_exception(e)
            logger.info("sent")


def _load_incoming_message(message_data):
    return TransportMessage(
        message_data['event'],
        message_data['topic'],
        message_data['payload'],
        message_data.get('ref')
    )


def dump_outgoing_message(message):
    return {
        'event': message.event,
        'topic': message.topic,
        'ref': message.ref,
        'payload': message.payload
    }



import asyncio

import pytest

from chunnel.utils import get_unless_done, DONE


@pytest.mark.asyncio
async def test_get_unless_done_getting():
    queue = asyncio.Queue()
    await queue.put(1)
    await queue.put(2)
    future = asyncio.Future()
    assert await get_unless_done(queue.get(), future) == 1
    assert await get_unless_done(queue.get(), future) == 2


@pytest.mark.asyncio
async def test_get_unless_done_when_done():
    queue = asyncio.Queue()
    future = asyncio.Future()
    future.set_result(True)
    assert await get_unless_done(queue.get(), future) is DONE


@pytest.mark.asyncio
async def test_get_unless_done_when_done_and_queue():
    queue = asyncio.Queue()
    future = asyncio.Future()
    future.set_result(True)
    await queue.put(1)
    assert await get_unless_done(queue.get(), future) is DONE



from concurrent.futures import FIRST_COMPLETED
import asyncio

__all__ = ['DONE', 'get_unless_done']


class DONE():
    '''
    Singleton that indicates a DONE when returned from get_unless_done.
    '''
    pass


# TODO: Tests
async def get_unless_done(getter_future_or_coro, done_future):
    '''
    Wraps a get operation & a future that indicates when we should stop
    getting.

    If the done_future is resolved while waiting for the get, we cancel the get
    and return DONE.

    :params getter_task_or_coro:   A getter.
    :params done_future:           A future that indicates we are done.
    '''
    getter_future = asyncio.ensure_future(getter_future_or_coro)
    done, pending = await asyncio.wait(
        (getter_future, done_future),
        return_when=FIRST_COMPLETED
    )
    if done_future in done:
        if not getter_future.done():
            getter, = pending
            getter.cancel()

        return DONE

    return await getter_future



from unittest.mock import sentinel
import asyncio

import pytest

from chunnel.channel import ChannelJoinFailure, ChannelLeaveFailure
from chunnel.messages import ChannelEvents
from chunnel.transports import TransportMessage

from .shared import set_reply


@pytest.yield_fixture
def socket(socket, event_loop):
    event_loop.run_until_complete(socket.connect())
    yield socket
    event_loop.run_until_complete(socket.disconnect())


@pytest.fixture
def channel(socket):
    return socket.channel("test:lobby", {})


@pytest.mark.asyncio
async def test_join(socket, channel):
    response, _ = await asyncio.gather(
        channel.join(),
        set_reply(
            socket,
            None,
            {'status': 'ok', 'response': sentinel.response}
        )
    )
    assert response == sentinel.response


@pytest.mark.asyncio
async def test_join_failure(socket, channel):
    with pytest.raises(ChannelJoinFailure):
        await asyncio.gather(
            channel.join(),
            set_reply(socket, None, {'status': 'error'})
        )


@pytest.mark.asyncio
async def test_leave(socket, channel):
    response, _ = await asyncio.gather(
        channel.join(),
        set_reply(socket, None, {'status': 'ok', 'response': {}})
    )
    await asyncio.gather(
        channel.leave(),
        set_reply(socket, None, {'status': 'ok'})
    )


@pytest.mark.asyncio
async def test_leave_failure(socket, channel):
    response, _ = await asyncio.gather(
        channel.join(),
        set_reply(socket, None, {'status': 'ok', 'response': {}})
    )
    with pytest.raises(ChannelLeaveFailure):
        await asyncio.gather(
            channel.leave(),
            set_reply(socket, None, {'status': 'error'})
        )


@pytest.mark.asyncio
async def test_context_manager_join(socket):
    asyncio.ensure_future(
        set_reply(
            socket, None, {'status': 'ok', 'response': sentinel.response}
        )
    )
    # TODO: Not sure about this api...
    async with socket.channel("test:lobby", {}) as (channel, response):
        assert response == sentinel.response
        assert channel == socket.channels['test:lobby']
        asyncio.ensure_future(
            set_reply(socket, None, {'status': 'ok'})
        )


@pytest.mark.asyncio
async def test_send_message(socket, channel):
    send_future = asyncio.ensure_future(
        channel.push(sentinel.event, sentinel.payload)
    )
    await asyncio.sleep(0)
    assert socket.transport.outgoing.qsize() == 1
    msg, sent_future = socket.transport.outgoing.get_nowait()
    assert msg.topic == channel.topic
    assert msg.event == sentinel.event
    assert msg.payload == sentinel.payload
    sent_future.set_result(True)

    assert await send_future


@pytest.mark.asyncio
async def test_message_replies(socket, channel):
    sent_message, _ = await asyncio.gather(
        channel.push(sentinel.event, sentinel.payload),
        set_reply(socket, None, {'status': 'ok', 'response': 'abcd'})
    )
    response = await sent_message.response()
    assert response == 'abcd'


@pytest.mark.asyncio
async def test_message_receive(socket, channel):
    message = TransportMessage(
        sentinel.event, channel.topic, sentinel.payload, sentinel.ref
    )
    await socket.transport.incoming.put(message)
    await socket.transport.incoming.put(message)
    message = await channel.receive()
    assert message.event == sentinel.event
    assert message.payload == sentinel.payload


@pytest.mark.asyncio
async def test_reply_to_received_message(socket, channel):
    await socket.transport.incoming.put(
        TransportMessage(
            sentinel.event, channel.topic, sentinel.payload, sentinel.ref
        )
    )
    incoming_message = await channel.receive()

    reply_future = asyncio.ensure_future(
        incoming_message.reply(sentinel.status, sentinel.response)
    )
    await asyncio.sleep(0)
    reply, sent_future = await socket.transport.outgoing.get()
    assert reply.topic == channel.topic
    assert reply.event == ChannelEvents.reply.value
    assert reply.payload == {
        'status': sentinel.status, 'response': sentinel.response
    }
    assert reply.ref == sentinel.ref

    sent_future.set_result(True)
    assert reply_future



from concurrent.futures import FIRST_COMPLETED
from urllib.parse import urlsplit
import asyncio
import logging

from .transports import (
    WebsocketTransport, TransportMessage, OutgoingTransportMessage
)
from .channel import Channel
from .messages import SentMessage, ChannelEvents, IncomingMessage
from .utils import get_unless_done, DONE

__all__ = ['Socket']

logger = logging.getLogger(__name__)


# TODO: Should this be called Socket? Dunno if it matches up with phoenix too
# well..
class Socket:
    '''
    A connection to a phoenix server.

    A transport will automatically be selected based on the URL provided.
    See the TRANSPORTS dict for more details.

    :param url:    The URL of the phoenix server to connect to.
    :param params: Optional parameters to use when connecting.
    '''

    # A mapping of url scheme -> transport.
    TRANSPORTS = {
        'ws': WebsocketTransport,
        'wss': WebsocketTransport
    }

    # TODO: Should these parameters be passed to connect?  Maybe not..
    def __init__(self, url, params):
        self.url = url
        self.params = params
        self.connected = False
        self.channels = {}
        self._incoming = asyncio.Queue()
        self._outgoing = asyncio.Queue()
        self._ref = 1
        self._response_futures = {}

    async def connect(self):
        if self.connected:
            raise Exception("Already connected!")

        transport_class = self.TRANSPORTS[urlsplit(self.url).scheme]
        self.transport = transport_class(
            self.url, self.params, self._incoming, self._outgoing
        )
        transport_task = asyncio.ensure_future(self.transport.run())

        await self.transport.ready

        self._transport_task = transport_task
        self._done_recv = asyncio.Future()
        self._recv_task = asyncio.ensure_future(self._recv_loop())
        # TODO: Ok, so this is cool - but how to tell if our transport_task has
        # failed.
        self.connected = True

    async def disconnect(self):
        if not self.connected:
            raise Exception("Not connected!")

        self._done_recv.set_result(True)
        await self.transport.stop()

        await asyncio.gather(self._recv_task, self._transport_task)

        self.connected = False

    def channel(self, topic, params):
        # TODO: What to do if we already have this channel?
        channel = Channel(self, topic, params)
        self.channels[topic] = channel
        return channel

    async def __aenter__(self):
        await self.connect()
        return self

    async def __aexit__(self, exc_type, exc, tb):
        await self.disconnect()

    async def _check_transport(self):
        # TODO: More thought around this function...
        if not self.connected:
            raise Exception("Not connected!")
        if self._transport_task.done():
            # We've probably excepted.
            # TODO: Do something more thorough here...
            self._transport_task.result()

    # TODO: _push_message?
    async def _send_message(self, topic, event, payload, ref=None):
        '''
        Sends a message to the remote.

        :param topic:   The topic to send the message on.
        :param event:   The name of the event to send.
        :param payload: The payload of the event.
        :param ref:     Optional ref to use for sending.
        :returns:       The ref of the event, which can be used to receive
                        replies.
        '''
        if not ref:
            ref = self._ref
            self._ref += 1

        message = OutgoingTransportMessage(
            TransportMessage(event, topic, payload, ref),
            asyncio.Future()
        )
        # TODO: add a done callback to reply_future that deletes it from
        # self._response_futures after a certain time...
        resp_future = asyncio.Future()
        self._response_futures[ref] = resp_future
        await self._outgoing.put(message)
        await message.sent
        # TODO: Return something slightly different....
        return SentMessage(resp_future)

    async def _recv_loop(self):
        '''
        Runs the socket receive loop.

        This will read incoming messages off the queue and attempt to route
        them to an appropriate place.
        '''
        while True:
            message = await get_unless_done(
                self._incoming.get(), self._done_recv
            )
            if message is DONE:
                break
            # TODO: Definitely need to handle phx_close...
            if message.event == ChannelEvents.reply.value:
                if message.payload['status'] == 'ok':
                    self._response_futures[message.ref].set_result(
                        message.payload.get('response')
                    )
                else:
                    # TODO: we can do better than this...
                    self._response_futures[message.ref].set_exception(
                        Exception("Response not ok!")
                    )
            else:
                channel = self.channels.get(message.topic)
                if channel:
                    await channel._incoming_messages.put(
                        IncomingMessage(message, self)
                    )



import asyncio
from collections import namedtuple


TransportMessage = namedtuple(
    'TransportMessage', ['event', 'topic', 'payload', 'ref']
)

# TODO: Could maybe call this Push to mirror it's name in phoenix js imp.
OutgoingTransportMessage = namedtuple(
    'OutgoingTransportMessage', ['message', 'sent']
)


class BaseTransport:
    '''
    The base class for a transport.

    Transports are used to implement the sending & receiving of messages in
    chunnel. Each transport is constructed with 2 queues - a queue for incoming
    messages and a queue for outgoing messages.

    The incoming message queue should contain TransportMessage namedtuples.

    The outgoing message queue should contain OutgoingTransportMessage
    namedtuples.

    The transport should read messages from the outgoing queue and send them
    onto a phoenix server, and put any incoming messages onto the incoming
    queue.

    Transports are not responsible for interpreting the messages in any way,
    they just handle the communication.
    '''
    def __init__(self, incoming_queue, outgoing_queue):
        self.incoming = incoming_queue
        self.outgoing = outgoing_queue
        self.ready = asyncio.Future()

    async def run(self):
        '''
        Connects the transport and runs it's main loop.

        Will resolve `self.ready` when a connection has been made.
        '''
        raise NotImplementedError

    async def stop(self):
        '''
        Signals to the transport that it should stop.
        '''
        raise NotImplementedError



import asyncio

from chunnel.messages import IncomingMessage
from chunnel.transports.base import BaseTransport, TransportMessage


class TestTransport(BaseTransport):
    '''
    A transport implementation for testing.
    '''
    RESOLVE_READY = True

    def __init__(self, url, params, incoming, outgoing):
        self.url = url
        self.params = params
        super().__init__(incoming, outgoing)
        self._future = asyncio.Future()
        if self.RESOLVE_READY:
            self.ready.set_result(True)

    async def run(self):
        self.running = True
        await self._future
        self.running = False

    async def stop(self):
        self._future.set_result(True)


# TODO: Decide if this class is worth it.
# Currently just used in one place, and quite easy to do a gather w/
# set_reply instead...
class TestSender():
    '''
    Helper class for sending test messages and setting their responses.

    Example:

        sender = TestSender('test:topic', 'an_event', {})
        sender.set_reply({'whatever': 'you_want'})
        msg = await sender.send(socket)
        response = await msg.response()
        assert response == {'whatever': 'you_want'}
    '''
    def __init__(self, topic, event, payload):
        self._outgoing = {
            'topic': topic,
            'event': event,
            'payload': payload
        }

    def set_reply(self, payload):
        self._reply_payload = payload

    async def send(self, socket):
        if self._reply_payload:
            other_future = set_reply(
                socket, self._outgoing['topic'], self._reply_payload
            )
        else:
            other_future = _mark_sent(socket)

        _, sent_message = await asyncio.gather(
            other_future, socket._send_message(**self._outgoing)
        )
        return sent_message


async def _mark_sent(socket):
    '''
    Marks the next outgoing message as sent.
    '''
    msg = await socket.transport.outgoing.get()
    msg.sent.set_result(True)


async def set_reply(socket, topic, payload):
    '''
    Sets the reply to the next sent message.

    Also marks the message as sent if it is not already.
    '''
    msg = await socket.transport.outgoing.get()
    if not msg.sent.done():
        msg.sent.set_result(True)

    await socket.transport.incoming.put(
        TransportMessage('phx_reply', topic, payload, msg.message.ref)
    )



from collections import namedtuple
from enum import Enum


class MessageStatus(Enum):
    ok = 'ok'
    error = 'error'


class IncomingMessage:
    def __init__(self, transport_message, socket):
        self._transport_message = transport_message
        self._socket = socket

    @property
    def event(self):
        return self._transport_message.event

    @property
    def payload(self):
        return self._transport_message.payload

    async def reply(self, status, response):
        await self._socket._send_message(
            self._transport_message.topic,
            ChannelEvents.reply.value,
            {'status': status, 'response': response},
            self._transport_message.ref
        )


# TODO: Think about where this belongs...
class ChannelEvents(Enum):
    close = "phx_close",
    error = "phx_error"
    join = "phx_join"
    reply = "phx_reply"
    leave = "phx_leave"


# TODO: PushedMessage?
class SentMessage:
    def __init__(self, response_future):
        self._response_future = response_future

    async def response(self):
        # TODO: Definitely need to think more about timeouts...
        # Currently the self._future is wrapped in a timeout, but is that what
        # I want?
        resp = await self._response_future
        return resp



from unittest.mock import sentinel
import asyncio

import pytest

from chunnel.transports import TransportMessage

from .shared import TestTransport, TestSender


@pytest.mark.asyncio
async def test_connecting(socket, mocker):
    mocker.patch.object(TestTransport, 'RESOLVE_READY', new=False)

    connect_future = asyncio.ensure_future(socket.connect())
    await asyncio.sleep(0)
    assert socket.transport.params == sentinel.connect_params

    assert socket.transport
    assert socket.transport.url == 'ws://localhost'
    assert not connect_future.done()

    socket.transport.ready.set_result(True)
    await asyncio.sleep(0)
    assert connect_future.done()
    assert socket.connected
    await connect_future

    await socket.disconnect()


@pytest.mark.asyncio
async def test_connect_as_context_manager(socket):
    async with socket as socket:
        assert socket.connected

    assert not socket.connected


@pytest.mark.asyncio
async def test_creating_channel(socket):
    async with socket:
        channel = socket.channel("test:topic", sentinel.channel_params)
        assert channel.topic == 'test:topic'
        assert channel.params == sentinel.channel_params
        assert 'test:topic' in socket.channels
        assert socket.channels['test:topic'] == channel


@pytest.mark.asyncio
async def test_sending_message(socket):
    async with socket:
        send_future = asyncio.ensure_future(
            socket._send_message(
                sentinel.topic, sentinel.event, sentinel.payload
            )
        )
        await asyncio.sleep(0)
        assert socket.transport.outgoing.qsize() == 1
        message, sent_future = socket.transport.outgoing.get_nowait()
        assert message.event == sentinel.event
        assert message.topic == sentinel.topic
        assert message.payload == sentinel.payload
        assert message.ref

        sent_future.set_result(True)

        await asyncio.sleep(0)
        assert send_future.done()
        assert await send_future


@pytest.mark.asyncio
async def test_replies_routed_correctly(socket):
    async with socket:
        sender = TestSender(sentinel.topic, sentinel.event, sentinel.payload)
        sender.set_reply({'status': 'ok', 'response': sentinel.response})

        sent_message = await sender.send(socket)

        assert await sent_message.response() == sentinel.response


@pytest.mark.asyncio
async def test_channel_messages_routed_correctly(socket):
    async with socket:
        channel = socket.channel("test:topic", sentinel.channel_params)
        message = TransportMessage(
            sentinel.event, "test:topic", sentinel.payload, sentinel.ref
        )
        await socket.transport.incoming.put(message)
        message = await channel.receive()
        assert message.event == sentinel.event
        assert message.payload == sentinel.payload



from unittest.mock import sentinel
import asyncio
import logging

import pytest

from chunnel.socket import Socket

from .shared import TestTransport


@pytest.fixture(autouse=True)
def setup_logging():
    logging.basicConfig(level=logging.DEBUG)


@pytest.yield_fixture
def event_loop():
    """
    Create an instance of the default event loop for each test case.

    This implementation is mostly a workaround for pytest-asyncio issues #29 &
    #30
    """
    policy = asyncio.get_event_loop_policy()
    res = policy.new_event_loop()
    asyncio.set_event_loop(res)
    res._close = res.close
    res.close = lambda: None

    yield res

    res._close()


@pytest.fixture
def socket(mocker, event_loop):
    mocker.patch.dict(Socket.TRANSPORTS, {'ws': TestTransport})
    return Socket('ws://localhost', sentinel.connect_params)



#!/usr/bin/env python

from setuptools import setup, find_packages

REQUIREMENTS = ['websockets']
LONG_DESCRIPTION = '''
Chunnel
-----

A python client for phoenix channels.

See the README at https://github.com/obmarg/chunnel for more details.

'''

setup(
    name='chunnel',
    version='0.1.0',
    url='https://github.com/obmarg/chunnel',
    description='Phoenix channels client library',
    long_description=LONG_DESCRIPTION,
    author='Graeme Coupar',
    author_email='grambo@grambo.me.uk',
    packages=find_packages(exclude=['tests']),
    install_requires=REQUIREMENTS,
    zip_safe=False,
    include_package_data=True,
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.5'
    ]
)



from uuid import uuid4
import asyncio
import os

import pytest
import requests

from chunnel.socket import Socket

TEST_CARD_URL = os.getenv('TEST_CARD_URL')
SKIP_TESTS = TEST_CARD_URL is None


@pytest.fixture
def user_id():
    id_ = str(uuid4())
    response = requests.post(
        'http://{}/api/users'.format(TEST_CARD_URL),
        json={"user": {"id": id_, "rooms": ["lobby"]}}
    )
    assert response.status_code == 201
    return id_


def make_socket(user_id):
    return Socket(
        'ws://{}/socket/websocket'.format(TEST_CARD_URL),
        {'user_id': user_id}
    )


@pytest.fixture
def socket(event_loop, user_id):
    return make_socket(user_id)


@pytest.fixture
def socket2(event_loop, user_id):
    return make_socket(user_id)


@pytest.mark.skipif(SKIP_TESTS, reason="TEST_CARD_URL env var not set")
@pytest.mark.asyncio
async def test_join_and_ping(socket):
    async with socket:
        channel = socket.channel("room:lobby", {'join': 'params'})
        assert await channel.join() == {'join': 'params'}
        ping = await channel.push("ping", {"some": "data"})
        assert await ping.response() == {"some": "data"}


@pytest.mark.skipif(SKIP_TESTS, reason="TEST_CARD_URL env var not set")
@pytest.mark.asyncio
async def test_join_and_shout(socket, socket2):
    async with socket:
        async with socket2:
            socket1_channel = socket.channel("room:lobby", {})
            socket2_channel = socket2.channel("room:lobby", {})
            await asyncio.gather(
                socket1_channel.join(), socket2_channel.join()
            )
            shout_payload = {"hello": "is anybody there?"}
            await socket1_channel.push("shout", shout_payload)

            incoming = await socket1_channel.receive()
            assert incoming.payload == shout_payload

            incoming = await socket2_channel.receive()
            assert incoming.payload == shout_payload
